#!/usr/bin/env python
# -*- coding: utf-8 -*-
#*******************************************************************************
#**  This file is part of HgWeb Manager.                                     ***
#**                                                                          ***
#**  Copyright (C) 2014                                                      ***
#**  CodeGNU Solutions <Licensing _AT_ CodeGNU _DOT_ com>                    ***
#**                                                                          ***
#**  This program is free software: you can redistribute it and/or modify it ***
#**  under the terms of the GNU Affero General Public License as published   ***
#**  by the Free Software Foundation, either version 3 of the License, or    ***
#**  (at your option) any later version.                                     ***
#**                                                                          ***
#**  This program is distributed in the hope that it will be useful, but     ***
#**  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
#**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
#**  See the GNU Affero General Public License for more details.             ***
#**                                                                          ***
#**  You should have received a copy of the GNU Affero General Public License***
#**  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
#*******************************************************************************

"""
 IMPORTS
"""
from hwm import settings


def create(repository):
	""" Creates a new repository under HgWeb.

	 @pre The repository to create cannot already exist.

	 @param[in] repository	The targeted repository to create.

	 @post The file system now contains a new directory in the repository path
	  that is the newly generated repository. The new repository is registered
	  in HgWeb to display and is secured to prevent accidental modification.

	 @return When the creation is successful gives true, else-wise false.
	"""
	import sys
	import os
	import shutil

	print "Adding repository: %s" % repository.StorageName
	if __doesRepositoryExist(repository):
		print >>sys.stderr, "The repository '%s' already exists." % repository.StorageName
		return False

	# Create the requested repository.
	status = __addRepository(repository)
	if 0 == status:
		print "Initialised repository directory."
	else:
		print >>sys.stderr, "Failed to create repository directory. OS Error Code:", status
		return False

	# Fill out the details of about the repository.
	if __writeRepositoryDetails(repository):
		print "Successfully wrote details."
	else:
		print >>sys.stderr, "Failed to write details."
		os.chmod(settings.RepositoryPath + os.sep + repository.StorageName + os.sep + '.hg' + os.sep + "hgrc", stat.S_IWRITE)
		shutil.rmtree(settings.RepositoryPath + os.sep + repository.StorageName)
		return False

	if not register(repository):
		os.chmod(settings.RepositoryPath + os.sep + repository.StorageName + os.sep + '.hg' + os.sep + "hgrc", stat.S_IWRITE)
		shutil.rmtree(settings.RepositoryPath + os.sep + repository.StorageName)
		return False
	else:
		print 'Repository added.'
		return True

def register(repository):
	""" Connects an existing repository to the mercurial web registry.

	 @pre The repository to register must already exist.

	 @param[in] repository	The targeted repository to register.

	 @post The new repository is registered in HgWeb to display and is
	  secured to prevent accidental modification.

	 @return When the registration is successful gives true, else-wise false.
	"""
	import sys

	print "Registering the repository: %s" % repository.StorageName
	if not __doesRepositoryExist(repository):
		print >>sys.stderr, "The repository '%s' does not exists." % repository.StorageName
		return False

	# Add the new repository to the web registry.
	if __registerRepository(repository):
		print "Successfully registered repository."
		return True
	else:
		print >>sys.stderr, "Failed to register repository."
		return False

def modify(newRepository, currentStorageName):
	""" Changes an existing repository in HgWeb.

	 @pre The repository to modify must already exist.

	 @param[in] newRepository		The new values for the repository to
								 modify.
	 @param[in] currentStorageName	The current storage name of the
								 repository to modify.

	 @post The repository is changed with the requested modifications.

	 @return When the modification is successful gives true, else-wise false.
	"""
	import sys
	from manrepo import Repository

	print "Modifying repository: " + currentStorageName
	oldRepository = Repository(currentStorageName)

	if not __doesRepositoryExist(oldRepository):
		print >>sys.stderr, "The repository '%s' was not found." % oldRepository.StorageName
		return False

	# Remove the current repository from the web registry.
	if not __unregisterRepository(oldRepository):
		print >>sys.stderr, "Failed to unregister repository."
		return False

	# Update renamed repositories
	if not oldRepository == newRepository:
		if __renameStorage(oldRepository, newRepository.StorageName):
			print "Renamed repository %s to %s." % (currentStorageName, newRepository.StorageName)
		else:
			print >>sys.stderr, "Failed to rename repository."
			if __registerRepository(oldRepository):
				print "Successfully rolled backed action."
			else:
				print >>sys.stderr, "Failed to roll back action."
			return False

	# Fill out the details of about the new repository.
	if __writeRepositoryDetails(newRepository):
		print "Successfully wrote new details."
	else:
		print >>sys.stderr, "Failed to write new details."
		if __registerRepository(oldRepository):
			print "Successfully rolled backed action."
		else:
			print >>sys.stderr, "Failed to roll back action."
		return False

	# Add the new repository to the web registry.
	if __registerRepository(newRepository):
		print 'Repository modified.'
		return True
	else:
		print >>sys.stderr, "Failed to register repository."
		return False

def delete(repository):
	""" Removes an existing repository.

	 @pre The repository to remove must already exist.

	 @param[in] repository	The targeted repository to delete.

	 @post The file system no longer contains a directory in the repository
	  path that is the supplied repository. The repository is no longer
	  registered in HgWeb.

	 @return When the deletion is successful gives true, else-wise false.
	"""
	import sys

	if not __doesRepositoryExist(repository):
		print >>sys.stderr, "The repository %s was not found." % repository.StorageName
		return False

	print "Removing repository: %s" % repository.StorageName
	if __unregisterRepository(repository):
		print "Successfully unregistered repository."
	else:
		print >>sys.stderr, "Failed to unregister repository."
		return False

	if __removeRepository(repository):
		print "Repository removed."
		return True
	else:
		print >>sys.stderr, "Failed to delete the repository."
		return False



#--- Functions for fulfilling repository management.



def __renameStorage(oldRepo, newName):
	""" Changes the storage name of an existing repository.

	 @param[in] oldRepo		The existing repository to whose storage name is
						 expected to be changed.
	 @param[in] newName		The new storage name for the repository.

	 @return Gives true when the storage name is successfully renamed,
	  else-wise gives false.
	"""
	import os
	import sys
	import shutil

	try:
		if oldRepo.StorageName == newName:
			return True
		elif os.exists(settings.RepositoryPath + os.sep + newName):
			print >>sys.stderr, "Cannot rename the existing repository %s because the name %s already exists." % (oldRepo.StorageName, newName)
			return False
		else:
			shutil.move(settings.RepositoryPath + os.sep + oldRepo.StorageName, settings.RepositoryPath + os.sep + newName)
	except OSError as e:
		print >>sys.stderr, "Rename error({0}): {1}".format(e.errno, e.strerror)
		return False
	return True

def __doesRepositoryExist(repo):
	""" Checks to see if a repository exists.

	 @param[in] repo	The targeted repository for which to check.

	 @return Gives true when the repository exists, false else-wise.
	"""
	import os

	if (repo is None) or (repo.StorageName is None):
		return False
	elif os.path.isdir(settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hg'):
		return True
	else:
		return False

def __unregisterRepository(repo):
	""" Removes the registration of the supplied repository.

	 @param[in] repo	The targeted repository for which to unregister.

	 @post The HgWeb configuration file does not contain the repository details
	  and is returned to a read-only state. Upon any error the standard error
	  buffer contains an error message.

	 @return When the process is successful gives true, else-wise false.
	"""
	import tempfile
	import os
	import shutil

	foundRepo=False
	tempFile=None
	with tempfile.NamedTemporaryFile(delete=False) as target:
		tempFile = target.name
		with open(settings.HgWebPath + os.sep + 'hgweb.config', 'r') as source:
			for line in source:
				if repo.StorageName in line:
					foundRepo=True
				else:
					target.write(line)
		target.flush()
		if not source.closed:
			while not source.closed:
				print "Waiting three seconds for configuration to close."
				time.sleep(3)
	try:
		shutil.copy2(tempFile, settings.HgWebPath + os.sep + 'hgweb.config')
	except IOError as e:
		print >>sys.stderr, "Unregister error({0}): {1}".format(e.errno, e.strerror)
		foundRepo = False
	os.remove(tempFile)
	return foundRepo

def __removeRepository(repo):
	""" Removes an existing Mercurial repository.

	 @pre The requested repository to delete must already exist.

	 @param[in] repo	The targeted repository for which to delete.

	 @post Upon any error the standard error buffer contains an error message.

	 @return When the process is successful gives true, else-wise false.
	"""
	import shutil
	import os
	import stat

	os.chmod(settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hg' + os.sep + "hgrc", stat.S_IWRITE)
	shutil.rmtree(settings.RepositoryPath + os.sep + repo.StorageName)
	return True

def __addRepository(repo):
	""" Initialises a new Mercurial repository.

	 @pre The requested repository to create cannot already exist.

	 @param[in] repo	The targeted repository for which to create.

	 @post Upon any error the standard error buffer contains an error message.

	 @return When the process is successful gives true, else-wise false.
	"""
	import subprocess
	import sys
	import os

	try:
		return subprocess.call([settings.HgCommand, "init", settings.RepositoryPath + os.sep + repo.StorageName])
	except OSError as e:
		print >>sys.stderr, "Initialisation error({0}): {1}".format(e.errno, e.strerror)
		return False
	return True

def __writeRepositoryDetails(repo):
	""" Writes the details of a repository to the repository's
	 configuration file.

	 @param[in] repo	The targeted repository for which to write the
					 configuration details.

	 @post The HgWeb configuration file contains the new repository details and
	  is returned to a read-only state.

	 @return When the process is successful gives true, else-wise false.
	"""
	import os
	import stat
	import sys

	config = settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hg' + os.sep + "hgrc"
	if os.path.exists(config):
		hgrcMode = os.stat(config)[stat.ST_MODE]
		os.chmod(config, hgrcMode | stat.S_IWRITE)
	else:
		hgrcMode = stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH

	with open(config, "w") as hgrc:
		hgrc.write("[web]\n")

		if repo.DisplayName is None:
			hgrc.write("name = " + repo.StorageName + '\n')
		else:
			hgrc.write("name = " + repo.DisplayName + '\n')

		if repo.Description is not None:
			hgrc.write("description = " + repo.Description + '\n')

		if repo.Contact is not None:
			hgrc.write("contact = " + repo.Contact + '\n')
	# Ensure the repository details isn't accidentally modified.
	os.chmod(config, hgrcMode & (~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH))

	# Ensure the repository isn't accidentally deleted, moved, or blocked.
	if "win32" == sys.platform:
		status = __guardWindowsRepository(repo)
		if 0 != status:
			exit(status)
	return True

def __guardWindowsRepository(repo):
	""" Guards the supplied repository when on a Windows system.

	 @param[in] repo	The targeted repository for which to secure.

	 @return The exit status value from securing the repository. The value will
	  be zero upon success, all other values indicate failure.
	"""
	import subprocess
	import os
	import sys

	status = subprocess.call(["attrib.exe", '+H', '+I', '+S', settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hg'])
	if 0 == status:
		print "Secured repository store."
	else:
		print >>sys.stderr, "Failed to secure repository store.", status
	return status

def __registerRepository(repo):
	""" Adds the registration details of the repository.

	 @param[in] repo	The targeted repository for which to register in HgWeb.

	 @post The HgWeb configuration file contains the new repository details and
	  is returned to a read-only state. Upon any error the standard error
	  buffer contains an error message.

	 @return When the process is successful gives true, else-wise false.
	"""
	import os
	import sys

	try:
		with open(settings.HgWebPath + os.sep + 'hgweb.config', 'a') as hgweb:
			hgweb.write('\n' + repo.StorageName + " = " + settings.RepositoryPath + os.sep + repo.StorageName)
		return True
	except IOError as e:
		print >>sys.stderr, "Web registration error({0}): {1}".format(e.errno, e.strerror)
		return False
